from experiments.learning import *
from lib.symbolic import super_integrate
from sympy import simplify

class FaultyMarbles(Configuration):
    def get_path(self):
        return "programs/other/faulty_marbles.st"
    
    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(1000, 10), iterations=300, seed=30),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 10), iterations=3000, seed=42),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 10), iterations=3000, seed=42),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 10), iterations=3000, seed=42),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(10000, 100, penalty_initial=10), iterations=10000, seed=35, detailed=True, save_best=True, save_best_criteria=1, smooth=False),
        ])
    
    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(3).get(dist)
        # _, template_vars, pre_expr, _, bias_vars = TwoLayerReluModel(3).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)
        return IsmProblemData.create(prog, dist, {"red": 5, "blue": 5, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=z3.And(convert_equation(guard, dist, set()), z3.Int("red_0") <= 5, z3.Int("blue_0") <= 5),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=False
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [5, 5])

class UnrealiableFaultyMarbles(Configuration):
    def get_path(self):
        return "programs/other/unreliable_faulty_marbles.st"
    
    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(1000, 10), iterations=300, seed=30),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 10), iterations=3000, seed=42),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 10), iterations=3000, seed=42),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 10), iterations=3000, seed=42),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(10000, 100, penalty_initial=10), iterations=10000, seed=35, detailed=True, save_best=True, save_best_criteria=1, smooth=False),
        ])
    
    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(3).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"red": 5, "blue": 5, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=z3.And(convert_equation(guard, dist, set()), z3.Int("red_0") <= 5, z3.Int("blue_0") <= 5),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=False
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [5, 5])

class SamplingFaultyMarbles(Configuration):
    def get_path(self):
        return "programs/other/faulty_marbles.st"

    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(1000, 10), iterations=300, seed=30),
            RefineDatasetStep(safe_cxex=20, unsafe_cxex=20),
            Run(loss=MeanBasedLoss(1000, 10), iterations=3000, seed=42),
            RefineDatasetStep(safe_cxex=20, unsafe_cxex=20),
            Run(loss=MeanBasedLoss(1000, 10), iterations=3000, seed=42, reuse_params=True),
            RefineDatasetStep(safe_cxex=20, unsafe_cxex=20),
            Run(loss=MeanBasedLoss(1000, 10), iterations=3000, seed=42, reuse_params=True),
            RefineDatasetStep(safe_cxex=20, unsafe_cxex=20),
            Run(loss=MeanBasedLoss(10000, 100), iterations=10000, seed=42, reuse_params=True, detailed=True, save_best=True, save_best_criteria=1, smooth=False),
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(3).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"red": 5, "blue": 5, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=z3.And(convert_equation(guard, dist, set()), z3.Int("red_0") <= 5, z3.Int("blue_0") <= 5),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=True
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [5, 5])

class SamplingUnreliableFaultyMarbles(Configuration):
    def get_path(self):
        return "programs/other/unreliable_faulty_marbles.st"

    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(1000, 10), iterations=300, seed=30),
            RefineDatasetStep(safe_cxex=20, unsafe_cxex=20),
            Run(loss=MeanBasedLoss(1000, 10), iterations=3000, seed=42),
            RefineDatasetStep(safe_cxex=20, unsafe_cxex=20),
            Run(loss=MeanBasedLoss(1000, 10), iterations=3000, seed=42, reuse_params=True),
            RefineDatasetStep(safe_cxex=20, unsafe_cxex=20),
            Run(loss=MeanBasedLoss(1000, 10), iterations=3000, seed=42, reuse_params=True),
            RefineDatasetStep(safe_cxex=20, unsafe_cxex=20),
            # Run(loss=MeanBasedLoss(1000, 10), iterations=3000, seed=42, reuse_params=True),
            # RefineDatasetStep(safe_cxex=20, unsafe_cxex=20),
            # Run(loss=MeanBasedLoss(1000, 10), iterations=3000, seed=42, reuse_params=True),
            # RefineDatasetStep(safe_cxex=20, unsafe_cxex=20),
            Run(loss=MeanBasedLoss(10000, 100), iterations=10000, seed=42, reuse_params=True, detailed=True, save_best=True, save_best_criteria=1, smooth=False),
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(3).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"red": 5, "blue": 5, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=z3.And(convert_equation(guard, dist, set()), z3.Int("red_0") <= 5, z3.Int("blue_0") <= 5),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=True
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [5, 5])

class FixedLoop(Configuration):
    def get_path(self):
        return "programs/assertion/fixed_loop.st"
 
    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(1000, 1), iterations=300, seed=30, detailed=False),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 1), iterations=5000, seed=42, detailed=False),
            Run(loss=MeanBasedLoss(1000, 1), iterations=6000, seed=57, reuse_params=True, detailed=True, save_best=True, save_best_criteria=1),
            Run(loss=MeanBasedLoss(1000, 1), iterations=2000, seed=73, reuse_params=True, smooth=False, detailed=True, learning_rate=5e-2, save_best=True, save_best_criteria=1)
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(1).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"t": 1, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=z3.And(convert_equation(guard, dist, set()), z3.Int("t_0") >= 1),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=False
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [])

class SamplingFixedLoop(Configuration):
    def get_path(self):
        return "programs/assertion/fixed_loop.st"
 
    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(1000, 1), iterations=300, seed=30, detailed=False),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 1), iterations=5000, seed=42, detailed=False),
            Run(loss=MeanBasedLoss(1000, 1), iterations=6000, seed=57, reuse_params=True, detailed=True, save_best=True, save_best_criteria=1),
            Run(loss=MeanBasedLoss(1000, 1), iterations=2000, seed=73, reuse_params=True, smooth=False, detailed=True, learning_rate=5e-2, save_best=True, save_best_criteria=1)
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(1).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"t": 1, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=z3.And(convert_equation(guard, dist, set()), z3.Int("t_0") >= 1),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=True
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [])

class FaultyLoop(Configuration):
    def get_path(self):
        return "programs/assertion/faulty_loop.st"
 
    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(1000, 1), iterations=300, seed=30, detailed=False),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 1), iterations=5000, seed=42, detailed=False),
            Run(loss=MeanBasedLoss(1000, 1), iterations=6000, seed=57, reuse_params=True, detailed=True, save_best=True, save_best_criteria=50),
            Run(loss=MeanBasedLoss(1000, 1), iterations=2000, seed=73, reuse_params=True, smooth=False, detailed=True, learning_rate=5e-2, save_best=True, save_best_criteria=50)
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(1).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"t": 1, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=z3.And(convert_equation(guard, dist, set()), z3.Int("t_0") >= 1),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=False
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [])

class SamplingFaultyLoop(Configuration):
    def get_path(self):
        return "programs/assertion/faulty_loop.st"
 
    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(1000, 1), iterations=300, seed=30, detailed=False),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 1), iterations=5000, seed=42, detailed=False),
            Run(loss=MeanBasedLoss(1000, 1), iterations=6000, seed=57, reuse_params=True, detailed=True, save_best=True, save_best_criteria=50),
            Run(loss=MeanBasedLoss(1000, 1), iterations=2000, seed=73, reuse_params=True, smooth=False, detailed=True, learning_rate=5e-2, save_best=True, save_best_criteria=50)
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(1).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"t": 1, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=z3.And(convert_equation(guard, dist, set()), z3.Int("t_0") >= 1),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=True
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [])
        
class FaultyEasy1Loop(Configuration):
    def get_path(self):
        return "programs/assertion/faulty_easy1.st"
 
    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(100, 1), iterations=300, seed=30, detailed=False),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(100, 1), iterations=5000, seed=42, detailed=False),           
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 10), iterations=6000, seed=57, reuse_params=True, detailed=True, save_best=True, save_best_criteria=1),
            Run(loss=MeanBasedLoss(1000, 100), iterations=2000, seed=73, reuse_params=True, smooth=False, detailed=True, learning_rate=5e-2, save_best=True, save_best_criteria=1)
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(1).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"x": 0, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=convert_equation(guard, dist, set()),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=False
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [])

class SamplingFaultyEasy1Loop(Configuration):
    def get_path(self):
        return "programs/assertion/faulty_easy1.st"
 
    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(100, 1), iterations=300, seed=30, detailed=False),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(100, 1), iterations=5000, seed=42, detailed=False),           
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 10), iterations=6000, seed=57, reuse_params=True, detailed=True, save_best=True, save_best_criteria=1),
            Run(loss=MeanBasedLoss(1000, 100), iterations=2000, seed=73, reuse_params=True, smooth=False, detailed=True, learning_rate=5e-2, save_best=True, save_best_criteria=1)
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(1).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"x": 0, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=convert_equation(guard, dist, set()),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=True
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [])

class FaultyNDecrLoop(Configuration):
    def get_path(self):
        return "programs/assertion/faulty_ndecr.st"
 
    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(100, 1), iterations=300, seed=30, detailed=False),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(100, 1), iterations=5000, seed=42, detailed=False),           
            Run(loss=MeanBasedLoss(1000, 10), iterations=6000, seed=57, reuse_params=True, detailed=True, save_best=True, save_best_criteria=1),
            Run(loss=MeanBasedLoss(1000, 10), iterations=2000, seed=73, reuse_params=True, smooth=False, detailed=True, learning_rate=5e-2, save_best=True, save_best_criteria=1)
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(1).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"i": 29, "n": 30, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=z3.And(convert_equation(guard, dist, set()), z3.Int("n_0") == 30),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=False
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [])

class SamplingFaultyNDecrLoop(Configuration):
    def get_path(self):
        return "programs/assertion/faulty_ndecr.st"
 
    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(100, 1), iterations=300, seed=30, detailed=False),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(100, 1), iterations=5000, seed=42, detailed=False),           
            Run(loss=MeanBasedLoss(1000, 10), iterations=6000, seed=57, reuse_params=True, detailed=True, save_best=True, save_best_criteria=1),
            Run(loss=MeanBasedLoss(1000, 10), iterations=2000, seed=73, reuse_params=True, smooth=False, detailed=True, learning_rate=5e-2, save_best=True, save_best_criteria=1)
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(1).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"i": 29, "n": 30, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=z3.And(convert_equation(guard, dist, set()), z3.Int("n_0") == 30),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=True
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [])

class FaultyLoopRare(Configuration):
    def get_path(self):
        return "programs/assertion/faulty_loop_rare.st"
 
    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(1000, 1), iterations=300, seed=30, detailed=False),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=500),
            Run(loss=MeanBasedLoss(1000, 1), iterations=5000, seed=42, detailed=False),
            Run(loss=MeanBasedLoss(1000, 1), iterations=6000, seed=57, reuse_params=True, detailed=True, save_best=True, save_best_criteria=50),
            Run(loss=MeanBasedLoss(1000, 1), iterations=2000, seed=73, reuse_params=True, smooth=False, detailed=True, learning_rate=5e-2, save_best=True, save_best_criteria=50)
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(1, with_bias=True).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"t": 1, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=z3.And(convert_equation(guard, dist, set()), z3.Int("t_0") >= 1),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=False
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [])

class SamplingFaultyLoopRare(Configuration):
    def get_path(self):
        return "programs/assertion/faulty_loop_rare.st"
 
    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(1000, 1), iterations=300, seed=30, detailed=False),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=500),
            Run(loss=MeanBasedLoss(1000, 1), iterations=5000, seed=42, detailed=False),
            Run(loss=MeanBasedLoss(1000, 1), iterations=6000, seed=57, reuse_params=True, detailed=True, save_best=True, save_best_criteria=50),
            Run(loss=MeanBasedLoss(1000, 1), iterations=2000, seed=73, reuse_params=True, smooth=False, detailed=True, learning_rate=5e-2, save_best=True, save_best_criteria=50)
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(1, with_bias=True).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"t": 1, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=z3.And(convert_equation(guard, dist, set()), z3.Int("t_0") >= 1),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=True
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [])

class FaultyLoopVarying(Configuration):
    def get_path(self):
        return "programs/assertion/faulty_loop_varying.st"

    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(1000, 10), iterations=300, seed=30, detailed=True),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 10), iterations=5000, seed=64, detailed=True, reuse_params=None, save_best=True, save_best_criteria=1),
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(2).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"t": 1, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=z3.And(z3.Int("t_0") >= 1, convert_equation(guard, dist, set())),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=False
        )
    
    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [])

class SamplingFaultyLoopVarying(Configuration):
    def get_path(self):
        return "programs/assertion/faulty_loop_varying.st"

    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(1000, 10), iterations=300, seed=30, detailed=True),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 10), iterations=5000, seed=64, detailed=True, reuse_params=None, save_best=True, save_best_criteria=1),
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(2).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"t": 1, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=z3.And(z3.Int("t_0") >= 1, convert_equation(guard, dist, set())),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=True
        )
    
    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [])

class Repulse(Configuration):
    def get_path(self):
        return "programs/other/repulse.st"
 
    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(100, 1), iterations=300, seed=30, detailed=False),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(100, 1), iterations=5000, seed=42, detailed=False),           
            Run(loss=MeanBasedLoss(100, 10), iterations=6000, seed=57, reuse_params=True, detailed=True, save_best=True, save_best_criteria=1),
            Run(loss=MeanBasedLoss(100, 100), iterations=2000, seed=73, reuse_params=True, smooth=False, detailed=True, learning_rate=5e-2, save_best=True, save_best_criteria=1)
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(1).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"x": 10},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("x_0") <= 100,
            unsafe_z3=z3.Int("x_0") >= 101,
            inv_z3=convert_equation(guard, dist, set()),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=False
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        def is_asserting(st):
            return st["x"] >= 101
        return build_real_dataset(prog, problem_data, [], is_asserting=is_asserting)

class SamplingRepulse(Configuration):
    def get_path(self):
        return "programs/other/repulse.st"
 
    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(100, 1), iterations=300, seed=30, detailed=False),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(100, 1), iterations=5000, seed=42, detailed=False),           
            Run(loss=MeanBasedLoss(100, 10), iterations=6000, seed=57, reuse_params=True, detailed=True, save_best=True, save_best_criteria=1),
            Run(loss=MeanBasedLoss(1000, 100), iterations=2000, seed=73, reuse_params=True, smooth=False, detailed=True, learning_rate=5e-2, save_best=True, save_best_criteria=1)
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(1).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"x": 10},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("x_0") <= 100,
            unsafe_z3=z3.Int("x_0") >= 101,
            inv_z3=convert_equation(guard, dist, set()),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=True
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        def is_asserting(st):
            return st["x"] >= 101
        return build_real_dataset(prog, problem_data, [], is_asserting=is_asserting)

class SamplingFaultyLoopUniform(Configuration):
    def get_path(self):
        return "programs/assertion/faulty_loop_uniform.st"
    
    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(1000, 1), iterations=300, seed=30, detailed=False),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 1), iterations=3000, seed=42, detailed=False),
            RefineDatasetStep(safe_cxex=200, unsafe_cxex=200),     
            Run(loss=MeanBasedLoss(1000, 10), iterations=3000, seed=57, detailed=True, reuse_params=True, save_best=True, save_best_criteria=1, smooth=False),
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(1).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.add_result(fix(pre_expr)) \
            .integrate_many([("p", 0), ("a", 2), ("result", 0)])
        post_expr = super_integrate(post_expr.pdf, Symbol("t_2"))

        return IsmProblemData.create(prog, dist, {"t": 1, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=z3.And(convert_equation(guard, dist, set()), z3.Real("t_0") >= 1),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=True
        )
    
    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [])

class SamplingFaultyRandomWalk(Configuration):
    def get_path(self):
        return "programs/assertion/faulty_random_walk.st"

    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(100, 10), iterations=300, seed=30, detailed=False),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(100, 10), iterations=3000, seed=42, detailed=False),
            RefineDatasetStep(safe_cxex=200, unsafe_cxex=200),    
            Run(loss=MeanBasedLoss(1000, 10), iterations=3000, seed=57, detailed=True, reuse_params=True, save_best=True, save_best_criteria=1, smooth=False),
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(1, with_bias=True).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.add_result(fix(pre_expr)) \
            .integrate_many([("p", 0), ("a", 2), ("result", 0)])
        post_expr = super_integrate(post_expr.pdf, Symbol("t_2"))

        return IsmProblemData.create(prog, dist, {"t": 1, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=convert_equation(guard, dist, set()),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=True
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [])

class SamplingRepulseUniform(Configuration):
    def get_path(self):
        return "programs/other/repulse_uniform.st"

    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(100, 10), iterations=300, seed=30, detailed=False),
            RefineDatasetStep(safe_cxex=0, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(100, 10), iterations=3000, seed=42, detailed=False),
            RefineDatasetStep(safe_cxex=10, unsafe_cxex=10),    
            Run(loss=MeanBasedLoss(1000, 100), iterations=3000, seed=57, detailed=False, reuse_params=True),
            RefineDatasetStep(safe_cxex=10, unsafe_cxex=10),    
            Run(loss=MeanBasedLoss(1000, 100), iterations=3000, seed=64, detailed=True, reuse_params=True, save_best=True, save_best_criteria=1, smooth=False),
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(1, with_bias=True).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.add_result(fix(pre_expr)) \
            .integrate_many([("x", 1), ("result", 0)])
        post_expr = super_integrate(post_expr.pdf, Symbol("q_0"))

        return IsmProblemData.create(prog, dist, {"x": 10},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Real("x_0") <= 100,
            unsafe_z3=z3.Real("x_0") >= 101,
            inv_z3=convert_equation(guard, dist, set()),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=True
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        def is_asserting(st):
            return st["x"] >= 101
        return build_real_dataset(prog, problem_data, [], is_asserting=is_asserting)

class Repulse2D(Configuration):
    def get_path(self):
        return "programs/other/repulse_2d.st"

    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(100, 10), iterations=300, seed=30, detailed=False),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 100), iterations=3000, seed=64, smooth=False),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 100), iterations=3000, seed=64, detailed=True, reuse_params=True, save_best=True, save_best_criteria=1, smooth=False),
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(2).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        pre_expr = soft_function(pre_expr)
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"x": 10, "y": 10},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Or(z3.Int("x_0") <= 100, z3.Int("y_0") <= 100),
            unsafe_z3=z3.And(z3.Int("x_0") > 100, z3.Int("y_0") > 100),
            inv_z3=convert_equation(guard, dist, set()),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=False
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        def is_asserting(st):
            return st["x"] > 100 and st["y"] > 100
        return build_real_dataset(prog, problem_data, [], is_asserting=is_asserting)

class SamplingRepulse2D(Configuration):
    def get_path(self):
        return "programs/other/repulse_2d.st"

    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(100, 10), iterations=300, seed=30, detailed=False),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 100), iterations=3000, seed=64, smooth=False),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 100), iterations=3000, seed=64, detailed=True, reuse_params=True, save_best=True, save_best_criteria=1, smooth=False),
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(2).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        pre_expr = soft_function(pre_expr)
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"x": 10, "y": 10},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Or(z3.Int("x_0") <= 100, z3.Int("y_0") <= 100),
            unsafe_z3=z3.And(z3.Int("x_0") > 100, z3.Int("y_0") > 100),
            inv_z3=convert_equation(guard, dist, set()),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=True
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        def is_asserting(st):
            return st["x"] > 100 and st["y"] > 100
        return build_real_dataset(prog, problem_data, [], is_asserting=is_asserting)

class FaultyLoopGaussian(Configuration):
    def get_path(self):
        return "programs/assertion/faulty_loop_gaussian.st"

    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(1000, 100), iterations=300, seed=30, detailed=True),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 100), iterations=3000, seed=64, detailed=True, reuse_params=None, save_best=True, save_best_criteria=1, smooth=False),
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = SosModel(1).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.add_result(fix(pre_expr))
        post_expr = post_expr.integrate_many([("p", 0), ("a", 2), ("result", 0)])
        post_expr = simplify((super_integrate(post_expr.pdf, Symbol("t_2"))))
        
        return IsmProblemData.create(prog, dist, {"t": 1, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=convert_equation(guard, dist, set()),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=False
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [])

class SamplingFaultyLoopGaussian(Configuration):
    def get_path(self):
        return "programs/assertion/faulty_loop_gaussian.st"

    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(1000, 100), iterations=300, seed=30, detailed=True),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 100), iterations=3000, seed=64, detailed=True, reuse_params=None, save_best=True, save_best_criteria=1, smooth=False),
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = SosModel(1).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.add_result(fix(pre_expr))
        post_expr = post_expr.integrate_many([("p", 0), ("a", 2), ("result", 0)])
        post_expr = simplify((super_integrate(post_expr.pdf, Symbol("t_2"))))
        
        return IsmProblemData.create(prog, dist, {"t": 1, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=convert_equation(guard, dist, set()),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=True
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [])

class RepsmRepulse(Configuration):
    def get_path(self):
        return "programs/other/repulse.st"

    def get_task(self):
        loss = MeanBasedLoss(100, 100, penalty_initial=10)
        return LearningTask(runs=[
            Run(loss=loss, iterations=10000, seed=30, detailed=True, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            # Run(loss=loss, iterations=10000, seed=64, detailed=True, learning_rate=1e-1),
            # RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            # Run(loss=loss, iterations=10000, seed=64, detailed=True, reuse_params=True, learning_rate=1e-1),
            # RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=loss, iterations=10000, seed=64, detailed=True, save_best=True, save_best_criteria=1, learning_rate=1e-1, smooth=False),
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(1, with_bias=True).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        pre_expr = -pre_expr
        post_expr = dist.get_post(pre_expr)

        return RepsmProblemData.create(prog, dist, {"x": 10},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("x_0") <= 100,
            unsafe_z3=z3.Int("x_0") >= 101,
            inv_z3=convert_equation(guard, dist, set()),
            non_increasing_cond_z3=convert_equation(fix(post_expr <= pre_expr - 1), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= 0), dist, template_vars_as_set),
            limit_expr=Abs(2 * Symbol("w_0,x"))
        )
    
    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        def is_asserting(st):
            return st["x"] >= 101
        return build_real_dataset(prog, problem_data, [], is_asserting=is_asserting)

class SamplingRepsmRepulse(Configuration):
    def get_path(self):
        return "programs/other/repulse.st"

    def get_task(self):
        loss = MeanBasedLoss(100, 100, penalty_initial=10)
        return LearningTask(runs=[
            Run(loss=loss, iterations=10000, seed=30, detailed=True, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            # Run(loss=loss, iterations=10000, seed=64, detailed=True, learning_rate=1e-1),
            # RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            # Run(loss=loss, iterations=10000, seed=64, detailed=True, reuse_params=True, learning_rate=1e-1),
            # RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=loss, iterations=10000, seed=64, detailed=True, save_best=True, reuse_params=True, save_best_criteria=1, learning_rate=1e-1),
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(1, with_bias=True).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        pre_expr = -pre_expr
        post_expr = dist.get_post(pre_expr)

        return RepsmProblemData.create(prog, dist, {"x": 10},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("x_0") <= 100,
            unsafe_z3=z3.Int("x_0") >= 101,
            inv_z3=convert_equation(guard, dist, set()),
            non_increasing_cond_z3=convert_equation(fix(post_expr <= pre_expr - 1), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= 0), dist, template_vars_as_set),
            limit_expr=None,
            use_samples=True
        )
    
    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        def is_asserting(st):
            return st["x"] >= 101
        return build_real_dataset(prog, problem_data, [], is_asserting=is_asserting)

# SCALE = 50, MAKE_SPARSE = True
class RepsmRepulse2D(Configuration):
    def get_path(self):
        return "programs/other/repulse_2d.st"

    def get_task(self):
        loss = MeanBasedLoss(1000, 10, penalty_initial=1)
        return LearningTask(runs=[
            Run(loss=loss, iterations=4000, seed=42, detailed=False, learning_rate=2e-2),
            RefineDatasetStep(safe_cxex=0, unsafe_cxex=100),
            Run(loss=loss, iterations=4000, seed=42, detailed=False, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=loss, iterations=10000, seed=42, detailed=False, reuse_params=True, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=loss, iterations=4000, seed=64, detailed=False, reuse_params=True, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            # Run(loss=loss, iterations=4000, seed=64, detailed=False, reuse_params=True, learning_rate=1e-1),
            # RefineDatasetStep(safe_cxex=200, unsafe_cxex=0),
            # Run(loss=loss, iterations=4000, seed=64, detailed=False, reuse_params=True, learning_rate=1e-1),
            # RefineDatasetStep(safe_cxex=200, unsafe_cxex=0),
            # Run(loss=loss, iterations=4000, seed=64, detailed=False, reuse_params=True, learning_rate=1e-1),
            # RefineDatasetStep(safe_cxex=200, unsafe_cxex=0),
            # Run(loss=loss, iterations=4000, seed=64, detailed=False, reuse_params=True, learning_rate=1e-1),
            # RefineDatasetStep(safe_cxex=200, unsafe_cxex=0),
            # Run(loss=loss, iterations=4000, seed=64, detailed=False, reuse_params=True, learning_rate=1e-1),
            # RefineDatasetStep(safe_cxex=0, unsafe_cxex=100),
            Run(loss=loss, iterations=3000, seed=64, detailed=True, reuse_params=True, save_best=True, save_best_criteria=1, learning_rate=2e-2),
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(2, with_bias=True).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        pre_expr = -pre_expr
        post_expr = dist.get_post(pre_expr)

        return RepsmProblemData.create(prog, dist, {"x": 10, "y": 10},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Or(z3.Int("x_0") <= 100, z3.Int("y_0") <= 100),
            unsafe_z3=z3.And(z3.Int("x_0") >= 101, z3.Int("y_0") >= 101),
            inv_z3=convert_equation(guard, dist, set()),
            non_increasing_cond_z3=convert_equation(fix(post_expr <= pre_expr - 1), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= 0), dist, template_vars_as_set),
            limit_expr=Max(Abs(2 * Symbol("w_0,x")) + Abs(2 * Symbol("w_1,x")),
              Abs(2 * Symbol("w_0,y")) + Abs(2 * Symbol("w_1,y")))
        )
    
    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        def is_asserting(st):
            return st["x"] >= 101 and st["y"] >= 101
        return build_real_dataset(prog, problem_data, [], is_asserting=is_asserting, count=1)

class FaultyLoopRegions(Configuration):
    def get_path(self):
        return "programs/assertion/faulty_loop_regions.st"

    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(1000, 10), iterations=300, seed=30),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 10), iterations=10000, seed=25),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 10), iterations=10000, seed=64, learning_rate=5e-2, reuse_params=False),
            Run(loss=MeanBasedLoss(1000, 10), iterations=10000, smooth=False, learning_rate=5e-3, seed=64, detailed=True, reuse_params=True, save_best=True, save_best_criteria=1),
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = TwoLayerReluModel(2).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"t": 10, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=z3.And(convert_equation(guard, dist, set())),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=False
        )
    
    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [])

class SamplingFaultyLoopRegions(Configuration):
    def get_path(self):
        return "programs/assertion/faulty_loop_regions.st"

    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(1000, 10), iterations=300, seed=30),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 10), iterations=10000, seed=25),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(1000, 10), iterations=10000, seed=64, learning_rate=5e-2, reuse_params=False),
            Run(loss=MeanBasedLoss(10000, 100), iterations=10000, smooth=False, learning_rate=5e-3, seed=64, detailed=True, reuse_params=True, save_best=True, save_best_criteria=1),
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = TwoLayerReluModel(2).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"t": 10, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=z3.And(convert_equation(guard, dist, set())),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=True
        )
    
    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [], count=1)

# SCALE = 50, MAKE_SPARSE = True, SAMPLE_COUNT = 10,000
class SamplingRepsmRepulse2D(Configuration):
    def get_path(self):
        return "programs/other/repulse_2d.st"

    def get_task(self):
        loss = MeanBasedLoss(1000, 10, penalty_initial=1)
        return LearningTask(runs=[
            Run(loss=loss, iterations=4000, seed=42, detailed=False, learning_rate=2e-2),
            RefineDatasetStep(safe_cxex=0, unsafe_cxex=100),
            Run(loss=loss, iterations=4000, seed=42, detailed=False, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=loss, iterations=10000, seed=42, detailed=False, reuse_params=True, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=loss, iterations=4000, seed=64, detailed=False, reuse_params=True, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=loss, iterations=3000, seed=64, detailed=True, reuse_params=True, save_best=True, save_best_criteria=1, learning_rate=2e-2),
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(2, with_bias=True).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        pre_expr = -pre_expr
        post_expr = dist.get_post(pre_expr)

        return RepsmProblemData.create(prog, dist, {"x": 10, "y": 10},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Or(z3.Int("x_0") <= 100, z3.Int("y_0") <= 100),
            unsafe_z3=z3.And(z3.Int("x_0") >= 101, z3.Int("y_0") >= 101),
            inv_z3=convert_equation(guard, dist, set()),
            non_increasing_cond_z3=convert_equation(fix(post_expr <= pre_expr - 1), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= 0), dist, template_vars_as_set),
            limit_expr=None,
            use_samples=True
        )
    
    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        def is_asserting(st):
            return st["x"] >= 101 and st["y"] >= 101
        return build_real_dataset(prog, problem_data, [], is_asserting=is_asserting, count=1)

class FaultyProbfact(Configuration):
    def get_path(self):
        return "programs/other/faulty_probfact.st"
 
    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(100, 1), iterations=300, seed=30, detailed=False),
            RefineDatasetStep(safe_cxex=0, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(100, 1), iterations=6000, seed=42, detailed=False),
            RefineDatasetStep(safe_cxex=0, unsafe_cxex=100),     
            Run(loss=MeanBasedLoss(100, 1), iterations=3000, seed=57, detailed=False, reuse_params=True),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=0),     
            Run(loss=MeanBasedLoss(500, 1), iterations=1000, seed=73, detailed=True, reuse_params=True, save_best=True, save_best_criteria=1)
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(1, with_bias=True).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        # pre_expr = soft_function(pre_expr)
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"i": 0, "s": 1, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=convert_equation(guard, dist, set()),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=False
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [])

# SAMPLE_COUNT = 10,000
class SamplingFaultyProbfact(Configuration):
    def get_path(self):
        return "programs/other/faulty_probfact.st"
 
    def get_task(self):
        return LearningTask(runs=[
            Run(loss=MeanBasedLoss(10, 1), iterations=300, seed=30, detailed=False),
            RefineDatasetStep(safe_cxex=0, unsafe_cxex=100),
            Run(loss=MeanBasedLoss(10, 1), iterations=6000, seed=42, detailed=False, learning_rate=5e-1),
            RefineDatasetStep(safe_cxex=0, unsafe_cxex=100),     
            Run(loss=MeanBasedLoss(10, 1), iterations=3000, seed=57, detailed=False, reuse_params=True),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=0),     
            Run(loss=MeanBasedLoss(500, 1), iterations=3000, seed=73, detailed=True, reuse_params=True, save_best=True, save_best_criteria=1)
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(1, with_bias=True).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        # pre_expr = soft_function(pre_expr)
        post_expr = dist.get_post(pre_expr)

        return IsmProblemData.create(prog, dist, {"i": 0, "s": 1, "a": 0},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Int("a_0") == 0,
            unsafe_z3=z3.Int("a_0") == 1,
            inv_z3=convert_equation(guard, dist, set()),
            non_increasing_cond_z3=convert_equation(fix(pre_expr >= post_expr), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= SCALE), dist, template_vars_as_set),
            use_samples=True
        )

    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        return build_real_dataset(prog, problem_data, [])

class RepsmPersist2D(Configuration):
    def get_path(self):
        return "programs/other/persist_2d.st"

    def get_task(self):
        loss = MeanBasedLoss(100, 100, penalty_initial=1)
        return LearningTask(runs=[
            Run(loss=loss, iterations=4000, seed=42, detailed=False, learning_rate=2e-2),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=loss, iterations=4000, seed=42, detailed=False, learning_rate=2e-2),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=0),
            Run(loss=loss, iterations=10000, seed=42, detailed=False, reuse_params=True, learning_rate=2e-2),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=0),
            Run(loss=loss, iterations=4000, seed=64, detailed=False, reuse_params=True, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=0),
            Run(loss=loss, iterations=4000, seed=64, detailed=False, reuse_params=True, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=0),
            Run(loss=loss, iterations=4000, seed=64, detailed=False, reuse_params=True, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=0),
            Run(loss=loss, iterations=4000, seed=64, detailed=False, reuse_params=True, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=0),
            Run(loss=loss, iterations=4000, seed=64, detailed=False, reuse_params=True, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=0),
            Run(loss=loss, iterations=4000, seed=64, detailed=False, reuse_params=True, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=0, unsafe_cxex=100),
            Run(loss=loss, iterations=3000, seed=64, detailed=True, reuse_params=True, save_best=True, save_best_criteria=1, learning_rate=2e-2),
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(2, with_bias=True).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        pre_expr = -pre_expr
        post_expr = dist.get_post(pre_expr)

        return RepsmProblemData.create(prog, dist, {"x": 10, "y": 10},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Or(z3.Int("x_0") <= 100, z3.Int("y_0") <= 100),
            unsafe_z3=z3.And(z3.Int("x_0") >= 101, z3.Int("y_0") >= 101),
            inv_z3=convert_equation(guard, dist, set()),
            non_increasing_cond_z3=convert_equation(fix(post_expr <= pre_expr - 1), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= 0), dist, template_vars_as_set),
            limit_expr=Max(Abs(2 * Symbol("w_0,x")) + Abs(2 * Symbol("w_1,x")),
              Abs(2 * Symbol("w_0,y")) + Abs(2 * Symbol("w_1,y")))
        )
    
    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        def is_asserting(st):
            return st["x"] >= 101 and st["y"] >= 101
        return build_real_dataset(prog, problem_data, [], is_asserting=is_asserting, count=1)

class SamplingRepsmPersist2D(Configuration):
    def get_path(self):
        return "programs/other/persist_2d.st"

    def get_task(self):
        loss = MeanBasedLoss(100, 100, penalty_initial=1)
        return LearningTask(runs=[
            Run(loss=loss, iterations=4000, seed=42, detailed=False, learning_rate=2e-2),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=100),
            Run(loss=loss, iterations=4000, seed=42, detailed=False, learning_rate=2e-2),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=0),
            Run(loss=loss, iterations=10000, seed=42, detailed=False, reuse_params=True, learning_rate=2e-2),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=0),
            Run(loss=loss, iterations=4000, seed=64, detailed=False, reuse_params=True, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=0),
            Run(loss=loss, iterations=4000, seed=64, detailed=False, reuse_params=True, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=0),
            Run(loss=loss, iterations=4000, seed=64, detailed=False, reuse_params=True, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=0),
            Run(loss=loss, iterations=4000, seed=64, detailed=False, reuse_params=True, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=0),
            Run(loss=loss, iterations=4000, seed=64, detailed=False, reuse_params=True, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=100, unsafe_cxex=0),
            Run(loss=loss, iterations=4000, seed=64, detailed=False, reuse_params=True, learning_rate=1e-1),
            RefineDatasetStep(safe_cxex=0, unsafe_cxex=100),
            Run(loss=loss, iterations=3000, seed=64, detailed=True, reuse_params=True, save_best=True, save_best_criteria=1, learning_rate=2e-2),
        ])

    def get_problem_data(self, prog: WhileLoopProgram):
        analyser = Analyser()
        guard, dist = analyser.analyse(prog)
        _, template_vars, pre_expr, _, bias_vars = ReluModel(2, with_bias=True).get(dist)
        template_vars_as_set = set(map(lambda x: x.name, template_vars))
        pre_expr = -pre_expr
        post_expr = dist.get_post(pre_expr)

        return RepsmProblemData.create(prog, dist, {"x": 10, "y": 10},
            pre_expr=pre_expr,
            post_expr=post_expr,
            template_vars=template_vars,
            bias_vars=bias_vars,
            safe_z3=z3.Or(z3.Int("x_0") <= 100, z3.Int("y_0") <= 100),
            unsafe_z3=z3.And(z3.Int("x_0") >= 101, z3.Int("y_0") >= 101),
            inv_z3=convert_equation(guard, dist, set()),
            non_increasing_cond_z3=convert_equation(fix(post_expr <= pre_expr - 1), dist, template_vars_as_set),
            indicating_cond_z3=convert_equation(fix(pre_expr >= 0), dist, template_vars_as_set),
            limit_expr=None,
            use_samples=True,
        )
    
    def get_dataset(self, prog: WhileLoopProgram, problem_data: IsmProblemData):
        def is_asserting(st):
            return st["x"] >= 101 and st["y"] >= 101
        return build_real_dataset(prog, problem_data, [], is_asserting=is_asserting, count=1)

def main():
    configuration = SamplingRepsmPersist2D()
    path, dataset, problem_data, task = configuration.get_all()
    experiment = LearningExperiment(task=task, tests=10)
    experiment.go(path, dataset, problem_data)

if __name__ == "__main__":
    jax.config.update("jax_platform_name", "gpu")
    print(jax.default_backend())
    main()
