from sympy import expand, symbols, Symbol, Ge, Le, Add, Mul, Rational, Integer, And, Lt, Gt, Eq
from fractions import Fraction
import pulp as pl

def convert_to_pulp(expr, get_var):
    if isinstance(expr, Rational):
        return Fraction(expr)
    elif expr.func == Symbol:
        return get_var(expr.name)
    elif expr.func == Add:
        result = 0
        for summand in expr.args:
            result += convert_to_pulp(summand, get_var)
        return result
    elif expr.func == Mul:
        result = 1
        for multiplicand in expr.args:
            result *= convert_to_pulp(multiplicand, get_var)
        return result
    print(expr)
    raise Exception("Conversion Error")

def convert(expr, state_vars):
    d = []
    c = {}
    args = []
    if expr.func == Add:
        args = expr.args
    else:
        args = [expr]
    for summand in args:
        if summand.func == Mul:
            found = None
            for factor in summand.args:
                if factor in state_vars:
                    found = factor
                    break
            if found != None:
                others = [factor for factor in summand.args if factor != found]
                c[found.name] = Mul(*others)
                continue
        elif summand in state_vars:
            c[summand.name] = Integer(1)
            continue
        d.append(summand)
    return c, d

def norm(constraint, state_vars):
    if constraint.func == Ge or constraint.func == Gt:
        constraint = Le(constraint.rhs, constraint.lhs)
    elif not (constraint.func == Le or constraint.func == Lt):
        print(constraint.func)
        raise Exception
    constraint = constraint.lhs - constraint.rhs
    return convert(expand(constraint), state_vars)

def transform_antecedents(antecedents):
    result = []
    for antecedent in antecedents:
        if antecedent.func == Eq:
            result.append(antecedent.lhs <= antecedent.rhs)
            result.append(antecedent.lhs >= antecedent.rhs)
        else:
            result.append(antecedent)
    return result

def farkas(conds, state_vars, get_var):
    model = pl.LpProblem()

    for i, (cond, antecedents_untransformed) in enumerate(conds):
        antecedents = transform_antecedents(antecedents_untransformed)
        c, d = norm(cond, state_vars)

        temp_vars = []
        for j in range(len(antecedents)):
            temp_vars.append(get_var(f"y_{i}_{j}"))
            model += temp_vars[j] >= 0

        for var in state_vars:
            target = 0
            if var.name in c:
                target = convert_to_pulp(c[var.name], get_var)

            total = 0
            for j, antecedent in enumerate(antecedents):
                a, _ = norm(antecedent, state_vars)
                val = Integer(0)
                if var.name in a:
                    val = a[var.name]
                total += convert_to_pulp(val, get_var) * temp_vars[j]

            model += total == target

        total = 0
        for j, antecedent in enumerate(antecedents):
            _, b = norm(antecedent, state_vars)
            total += convert_to_pulp(-Add(*b), get_var) * temp_vars[j]
        # print(cond)
        # print(d)
        model += total <= convert_to_pulp(-Add(*d), get_var)
    return model