from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Set, Tuple, Union
from fractions import Fraction

class PrimType(Enum):
    INT = 0
    REAL = 1
    BOOL = 2

@dataclass
class DistType:
    base: PrimType

Type = Union[PrimType, DistType]


@dataclass
class IntNode:
    value: int

@dataclass
class RealNode:
    value: Fraction

@dataclass
class BoolNode:
    value: bool

@dataclass
class IdentifierNode:
    name: str

@dataclass
class UnaryOperator:
    operator: str
    operand: "Expr"

@dataclass
class BinaryOperator:
    operator: str
    operand1: "Expr"
    operand2: "Expr"

@dataclass
class DistributionNode:
    operator: str
    args: List["Expr"]

Expr = Union[IntNode, RealNode, BoolNode, IdentifierNode, UnaryOperator, BinaryOperator, DistributionNode]

@dataclass
class AssignmentStatement:
    name: str
    expr: Expr

@dataclass
class BlockStatement:
    statements: List["Statement"]

@dataclass
class IfStatement:
    branches: List[Tuple[Expr, BlockStatement]]

@dataclass
class WhileStatement:
    guard: Expr
    body: BlockStatement

Statement = Union[AssignmentStatement, BlockStatement, IfStatement]

@dataclass
class Function:
    name: str
    args: List[Tuple[str, PrimType]]
    body: BlockStatement

@dataclass
class Program:
    functions: List[Function]

@dataclass
class WhileLoopProgram:
    init: List[Statement]
    args: List[Tuple[str, Type]]
    guard: Expr
    body: List[Statement]